(*:Mathematica Version: 3.0  *)

(*:Title:   The OPMatrix Package *)

(*:Author:  Liu Zhao 
            Inst. Mod. Phys., Northwest Univ., Xian 710069, China
            lzhao@phy.nwu.edu.cn
            
            Any comments, bug reports, suggestions etc welcomed.
*)

(*:Context: OPLinearAlgebra`OPMatrix`  *)

(*:Summary:
    This package is a collection of a number of functions aiming at linear algebraic 
    manipulations for operator-valued matrices. Ofcause orddinary matrices can 
    also be treated, but since the aim of of this package is for operator-valued
    objects, the treatment for ordinary matrices will be of very low efficiency.
    Therefore, it is strongly adviced that the users use functions for ordinary
    matrices whenever ordinary matrices are encountered.
    
    Summary of key functions:
    -- CircleTimes(DProd) -- Not previously implemented, defines direct products
       for matrices (applicable to any matrices, operator-valued or not)
    -- CirclePlus -- Not previouslu implemented, defines direct sum of matrices 
       (applicable to any matrices, operator-valued or not)
    -- Diamond -- Adding new rules to th Diamond function defined in OPScalar.m
       Now diamond recognizes matrices and it plays the roles of scalar product
       between a single-component operator and an operator-valued matrix, and that 
       of matrix multiplication (Dot) between two operator-valued matrices.
    -- OPInv -- Matrix inverse using Gaussian ellimination. This function works
       well for matrices whose entries are all symbolic, but may not work if the 
       matrix is numeric because it makes no pivot while doing the Gauss 
       ellimination.
    -- GaussDecompose -- Gauss decomposition of a matrix. This is actually the key 
       part of OPInv.
    -- GaussL, GaussD, GaussU -- These three functions gives the three parts of
       GaussDecompose. Since the symbolic Gauss Decomposition may be quite
       time consuming, so these three functions are implemented independently of 
       GaussDecompose, just in case only one or two parts of the result of GaussDecompose
       is needed in practice.
    -- LUFacOPMatrix -- This function is analogous to the addon package function
       LUFactorMatrices in the LinearAlgebra package. It works for operator-valued
       matrices.
    -- OPLinearSolve -- This function is analogous to the addon function LinearSolve.
       It solves the operator-valued linear systems of equations arranged in matrix 
       form A X = B or X A = B.
*)

(*:Keywords: Operator-valued matrices *)

(*:Requirements: OPLinearAlgebra`OPScalar`, 
   LinearAlgebra`MatrixManipulation` *)

(*:Package Version: 1.1 *)

(*:History:  Created in May, 2000 by Liu Zhao *)


BeginPackage["OPLinearAlgebra`OPMatrix`",
             "OPLinearAlgebra`OPScalar`",
             "LinearAlgebra`MatrixManipulation`"];

(* Unprotecting and clearing relevant objects *)

Unprotect[CircleTimes, DProd, Dot, Times, Diamond, Commutator, 
          NCPower, OPExpand, CirclePlus, CircleCommutator,
          RevDiamond,ComplementMatrix,LeftMinor,RightMinor,
          LeftDet,RightDet,ncinv,OPInv,DiagonalPart,
          LUFacOPMatrix, GaussL, GaussD, GaussU, GaussDecompose, 
          OPLinearSolve,NCPowerToDiamond,Tp,
          StrictUpperDiagonalMatrix,StrictLowerDiagonalMatrix];
Clear[DProd, CircleTimes, CirclePlus, CircleCommutator,
      RevDiamond,ComplementMatrix,LeftMinor,RightMinor,
      LeftDet,RightDet,ncinv,OPInv,DiagonalPart,
      LUFacOPMatrix, GaussL, GaussD, GaussU, GaussDecompose, 
      OPLinearSolve,NCPowerToDiamond,Tp,
      StrictUpperDiagonalMatrix,StrictLowerDiagonalMatrix];

(* Brief helps for new functions *)

DProd::usage =
"DProd[mat1,mat2,...] gves the succesive direct product of several 
matrices with operator-valued entries which is again a matrix." 

CircleTimes::usage=
"CircleTimes[mat1,mat2,...] gives the succesive direct product of several 
matrices with operator-valued entries which is again a matrix." 

CirclePlus::usage=
"CirclePlus[mat1,mat2,...] gives the direct sum of several square matrices."

CircleCommutator::usage=
"CircleCommutator[x ,y] is a combination of Commutator and CircleTimes. It gives a \
matrix whose number of rows (columns) is the number of rows (columns) for x times \
the number of rows (columns) of y, and whose entry at position \
{#rows(y)(i-1) +j, #columns(y)(k-1) +l} equals Commutator[x[[i, k]], y[[j, l]] \
(where of cause the range of {i, k} and {j, l} are restricted by the size of the \
matrices x and y."

RevDiamond::usage =
"RevDiamond[x] reverses the order of all Diamond subexpressions in a polynomial."

ComplementMatrix::usage=
"ComplementMatrix[i, j][x] gives a matrix which is x but with the i-th row \
and j-th column removed."

LeftMinor::usage=
"LeftMinor[i, j][x] gives the algebraic co-minor of x[[i, j]] in x with Diamond \
expressions arranged in left-to-right order."

RightMinor::usage=
"RightMinor[i, j][x] gives the algebraic co-minor of x[[i, j]] in x with Diamond \
expressions arranged in right-to-left order."

LeftDet::usage=
"LeftDet[x] gives the determinant of x with Diamond \
expressions arranged in left-to-right order."

RightDet::usage=
"RightDet[x] gives the determinant of x with Diamond \
expressions arranged in right-to-left order."

ncinv::usage=
"ncinv[x] gives the inverse of the matrix x of dimension bigger than 1. This is a \
naive implementation, which is very slow. Use OPInv instead."

OPInv::usage=
"OPInv[x, opts] gives the inverse of the matrix x. NCPower[x, -1] is \
automatically evaluated into OPInv[x] if x is of Gauss type. The option is \
TimeConstraint. Default is TimeConstraint -> 300 (seconds)."

DiagonalPart::usage=
"DiagonalPart[x] takes out the diagonal elements of x and form a new diagonal matrix \
of of them."

LUFacOPMatrix::usage=
"LUFacOPMatrix[x] performs LUFactorization for the operator matrix x. This is a \
clone of the function LUFactor in the LinearAlgebra`GaussianElimination` package, \
with minor modifications to make it work for operator matrices."

GaussL::usage=
"GaussL[x] gives the lower diagonal component in the Gauss decomposition of x."

GaussD::usage=
"GaussD[x] gives the diagonal component in the Gauss decomposition of x."

GaussU::usage=
"GaussU[x] gives the upper diagonal component in the Gauss decomposition of x."

GaussDecompose::usage=
"GaussDecompose[x] gives the list of three matrices, GaussL[x], GaussD[x] and \
GaussU[x]."

OPLinearSolve::usage=
"OPLinearSolve[List, Mat] solves the linear equation List == X . Mat; \
OPLinearSolve[Mat, List] solves the linear equation Mat . X == List."

NCPowerToDiamond::usage=
"NCPowerToDiamond[mat, integer] does the samething as PowerToDiamond \
does for matrices."

Tp::usage=
"Tp[x] gives the transpose of operator-valued expression x."

StrictUpperDiagonalMatrix::usage=
"StrictUpperDiagonalMatrix[x,n] gives a strict upper diagonal matrix, i.e. a matrix \
whose diagonal part is 1, lower triangular part is 0 and upper triangular part \
is x[i,j]."

StrictLowerDiagonalMatrix::usage=
"StrictLowerDiagonalMatrix[x,n] gives a strict lower diagonal matrix, i.e. a matrix \
whose diagonal part is 1, upper triangular part is 0 and lower triangular part \
is x[i,j]."




Begin["`Private`"];

(* adding matrix arguments to Diamond *)

refreshDiamond[]:=
Module[{},Unprotect[Diamond];
(* adding new rules *)
Diamond[c___, a_, M_?MatrixQ, b___] := 
    Diamond[c, Table[Diamond[a, M[[i, j]]], 
                    {i, 1, Length[M]}, {j, 1, Length[M]}
               ], b
    ] /; Not[MatrixQ[a]] && Head[a] =!= NCPower;
Diamond[c___, M_?MatrixQ, a_, b___] := 
    Diamond[c, Table[Diamond[M[[i, j]], a], 
                    {i, 1, Length[M]}, {j, 1, Length[M]}
               ], b 
    ] /; Not[MatrixQ[a]] && Head[a] =!= NCPower; 
Diamond[c___,x_?MatrixQ, y_?MatrixQ, b___] := 
    Block[{ret},Off[Thread::tdlen];
      ret=
       Diamond[c,
        Module[{tmp1, tmp2, rows, cols},
          tmp1 = Dimensions[x][[2]];
          tmp2 = Dimensions[y][[1]];
          rows = Dimensions[x][[1]];
          cols = Dimensions[y][[2]];
          If[tmp1 =!= tmp2, 
             Return[Message[Diamond::matrix]],
             Table[
               Sum[x[[i,k]] \[Diamond] y[[k,j]], {k, 1, tmp1}], 
               {i, 1, rows}, {j, 1, cols}
             ]
          ]
        ], b
       ];
      (*On[Thread::tdlen];*)
      ret
    ];
Protect[Diamond];
];

Diamond::matrix=
"Matrix size incompatible! Operation Failed!"

refreshDiamond[];

(* ExpandDiamondNCPower *)

ExpandDiamondNCPowerP[Literal[NCPower[x_?MatrixQ, n_]]] :=
    Which[Length[x] =!= Length[Transpose[x]], NCPower[x,n],
          True,  
             Module[{res,tmp,mat,dim,rules,ret},dim=Length[x];
                    If[SetAllOperatorsQ[], Null, SetOperators[mat]];
                    tmp=Table[mat[i,j],{i,1,dim},{j,1,dim}];
                    res=x;
                    rules=Flatten[Table[tmp[[i,j]]->x[[i,j]],{i,1,dim},{j,1,dim}]];
                    Do[res=Diamond[res,tmp] /. rules, {n-1}];
                    ret=Table[OPExpand[res[[i,j]]],{i,1,dim},{j,1,dim}];
                    If[SetAllOperatorsQ[], Null, ClearOperators[mat]];
                    ret
             ]
    ] /; IntegerQ[n] && Positive[n];
    
OPExpand[Literal[NCPower[x_, n_]]] := 
    Which[Length[x] <= 1 || Not[NumericQ[n]] || Not[OperatorQ[x]], NCPower[x, n],
          Positive[n], 
             If[IntegerQ[n], ExpandDiamondNCPowerP[NCPower[x, n]], NCPower[x, n]],
          Negative[n],
             If[Head[x] === Diamond,
                Block[{l, tmp}, l = Length[x];
                      Do[tmp[i] = OPExpand[NCPower[Part[x, l + 1 - i], n]], 
                         {i, 1, l}
                      ];
                      Apply[Diamond, Table[tmp[i], {i, 1, l}]] // OPExpand
                ],
                NCPower[x, n]
             ]
    ];


NCPower[x_, y_] := 
    NCPower[OPInv[x], -y] /; gaussQ[x] && (NumericQ[y] && Negative[y]);


(* Some more general linear algebraic functions for matrices *)

(* CircleTimes *)

CircleTimes[l_?MatrixQ,m_?MatrixQ] := BlockMatrix[Outer[Diamond,l,m]]

CircleTimes[l__?MatrixQ]:=Fold[CircleTimes, {{1}}, {l}]

DProd[l__?MatrixQ]:=CircleTimes[l]

(* CirclePlus *)

CirclePlusPlus[x_?MatrixQ] :=
    If[Length[x] =!= Length[Transpose[x]] , Message[CirclePlus::dim], x];
    
CirclePlus[x_?MatrixQ, y_?MatrixQ] :=
    If[Length[x] =!= Length[Transpose[x]] || 
       Length[y] =!= Length[Transpose[y]], 
       Message[CirclePlus::dim],
       Module[{tmp, dim1, dim2, zerom},
          dim1 = Length[x]; dim2 = Length[y];
          zerom = Table[0, {i, 1, dim1}, {j, 1, dim2}];
          BlockMatrix[{{x, zerom}, {Transpose[zerom], y}}]
       ]
    ];

CirclePlus[x_?MatrixQ, y___?MatrixQ] := Fold[CirclePlus, x, {y}];

CirclePlus::dim = "Failed to define direct sum for nonsquare matrices."

(* CircleCommutator *)

CircleCommutator[x_?MatrixQ,y_?MatrixQ] := BlockMatrix[Outer[Commutator,x,y]]

(* Commutator for matrices *)

Commutator[x_?MatrixQ,y_?MatrixQ] := 
    If[Length[x] =!= Length[Transpose[x]] || 
       Length[y] =!= Length[Transpose[y]] ||
       Length[x] =!= Length[y], Message[Diamond::matrix],
       x\[Diamond]y - y\[Diamond]x
    ];

(* Strict upper and lower diagonal matrices *)

StrictLowerDiagonalMatrix[fn_, size_Integer] :=
	Array[If[#1>=#2, If[#1>#2,fn[#1, #2],1], 0]&, {size, size}] /;
	size >= 0

StrictUpperDiagonalMatrix[fn_, size_Integer] :=
	Array[If[#1<=#2, If[#1<#2,fn[#1, #2],1], 0]&, {size, size}] /;
	size >= 0

(* ------------------------------------------------------------------------ *)
(* Noncommutative determinants and inverse *)

(* Matrix shape determination *)

UpperDiagonalMatrixQ[x_] := 
   Block[{l=Length[x]},
    If[Not[MatrixQ[x]] || l =!= Length[Transpose[x]], False, 
      x === x Table[If[i <= j, 1, 0], {i, 1, l}, {j, 1, l}]]
   ];
      
LowerDiagonalMatrixQ[x_] := 
   Block[{l=Length[x]},
    If[Not[MatrixQ[x]] || l =!= Length[Transpose[x]], False, 
      x === x Table[If[i >= j, 1, 0], {i, 1, l}, {j, 1, l}]]
   ];
      
DiagonalMatrixQ[x_] := 
   Block[{l=Length[x]},
    If[Not[MatrixQ[x]] || l =!= Length[Transpose[x]], False, 
      x === x Table[If[i == j, 1, 0], {i, 1, l}, {j, 1, l}]]
   ];
      
DiagonalPart[x_] := 
   Block[{l=Length[x]},
    If[Not[MatrixQ[x]] || l =!= Length[Transpose[x]], 
      Message[DiagonalPart::size], 
      x Table[If[i == j, 1, 0], {i, 1, l}, {j, 1, l}]]
   ];

DiagonalPart::size=
"DiagonalPart act only on square matrices."
      
StrictLowerDiagonalMatrixQ[x_] := 
    LowerDiagonalMatrixQ[x] && IdentityMatrix[Length[x]] === DiagonalPart[x];
    
StrictUpperDiagonalMatrixQ[x_] := 
    UpperDiagonalMatrixQ[x] && IdentityMatrix[Length[x]] === DiagonalPart[x];
    
gaussQ[x_]:= StrictLowerDiagonalMatrixQ[x] || StrictUpperDiagonalMatrixQ[x] ||
    DiagonalMatrixQ[x];

(* reverse the order of Diamond *)

rev[x_ + y_] := rev[x] + rev[y];
rev[x_] := rev[x] rev[y];
rev[x_Diamond] := Reverse[x];
rev[x_] := x /; Length[x] <= 1 ||
      (Head[x] =!= Plus && Head[x] =!= Diamond && Head[x] =!= Times) || 
      Not[OperatorQ[x]];
rev[-x_] := -rev[x];

RevDiamond[x_]:=rev[x];

(* All new codes *)

ComplementMatrix[i_, j_][x_?MatrixQ] := 
    Transpose[Drop[Transpose[Drop[x, {i}]], {j}]];

LeftMinor[i_,j_][x_?MatrixQ] := 
    (-1)^(i+j) LeftDet[ComplementMatrix[i,j][x]];
RightMinor[i_,j_][x_?MatrixQ] := 
    (-1)^(i+j) RightDet[ComplementMatrix[i,j][x]];
    
LeftDet[x_?MatrixQ] :=
    Module[{l}, l = Length[x];
      If[l=!=Length[Transpose[x]], Message[Det::mat],
         If[OperatorQ[Flatten[x]],
            If[l == 1, x[[1, 1]],
               Sum[(-1)^(k + l) x[[l, k]] \[Diamond] LeftDet[ComplementMatrix[l, k][x]], 
                   {k, 1, l}
               ] 
            ],
            Det[x]
         ]
      ]
    ] /; Not[StrictLowerDiagonalMatrixQ[x]] && Not[StrictUpperDiagonalMatrixQ[x]];

  
RightDet[x_?MatrixQ] :=
    Module[{l}, l = Length[x];
      If[l=!=Length[Transpose[x]], Message[Det::mat],
         If[OperatorQ[Flatten[x]],
            If[l == 1, x[[1, 1]],
               Sum[(-1)^(k + l) RightDet[ComplementMatrix[l, k][x]] \[Diamond] x[[l, k]], 
                   {k, 1, l}
               ]
            ],
            Det[x]
         ]
      ]
    ] /; Not[StrictLowerDiagonalMatrixQ[x]] && Not[StrictUpperDiagonalMatrixQ[x]];

LeftDet[x_?MatrixQ] :=  
    1 /; StrictLowerDiagonalMatrixQ[x] || StrictUpperDiagonalMatrixQ[x];

RightDet[x_?MatrixQ] := 
    1 /; StrictLowerDiagonalMatrixQ[x] || StrictUpperDiagonalMatrixQ[x];

(* naive inverse codes, slowest *)

upperncinv[x_?MatrixQ] := 
  Module[{tmp,l}, tmp = LeftDet[x]^-1; l=Length[x];
    Table[ RightMinor[j, i][x], {i, 1, l}, {j, 1, l}] \[Diamond] tmp
  ];

lowerncinv[x_?MatrixQ] := 
  Module[{tmp,l}, tmp = RightDet[x]^-1; l=Length[x];
    tmp \[Diamond] Table[LeftMinor[j, i][x], {i, 1, l}, {j, 1, l}]
  ];
  
ncinv[x_?MatrixQ] :=
  Which[StrictUpperDiagonalMatrixQ[x], upperncinv[x],
        StrictLowerDiagonalMatrixQ[x], lowerncinv[x],
        DiagonalMatrixQ[x],            
          DiagonalMatrix[Table[OPExpand[Power[x[[i,i]],-1]],{i,1,Length[x]}]],
        True,Message[inv::gauss]
  ];

inv::gauss:= "Unable to treat non-triangular matrices. Use OPInv instead."

(* fast codes for gauss type matrices *)

upperinv[A_] :=
    Which[
      Length[A] == 1, A,
      Length[A] == 2, A {{1, -1}, {1, 1}},
      True,
        Module[{mat, invmat, lastcol, lastrow, l, m}, l = Length[A];
          mat = ComplementMatrix[l, l][A];
          invmat = upperinv[mat];
          lastcol = 
            Transpose[{
              Table[-Sum[invmat[[k, i]] \[Diamond] A[[i, l]], {i, k, l - 1}], 
                   {k, 1, l - 1}]
            }];
          lastrow = {Join[Table[0, {l - 1}], {1}]};
          ret = AppendColumns[AppendRows[invmat, lastcol], lastrow];
          Return[ret]
        ]
    ];

lowerinv[x_] :=
    Which[
      Length[x] == 1, x, 
      Length[x] == 2, x {{1, 1}, {-1, 1}}, 
      True, 
        Module[{mat, invmat, lastcol, lastrow, l}, l = Length[x];
          mat = ComplementMatrix[l, l][x];
          invmat = lowerinv[mat];
          lastrow = {Table[-Sum[x[[l, i]] \[Diamond] invmat[[i, k]], {i, k, l - 1}], 
                     {k, 1, l - 1}]
                    };
          lastcol = Transpose[{Join[Table[0, {l - 1}], {1}]}];
          ret = AppendRows[AppendColumns[invmat, lastrow], lastcol];
          Return[ret]
        ]
    ];

OPInv[x_?MatrixQ, opts___?OptionQ] :=
  Which[Not[MatrixQ[x]] || (Length[x]=!=Length[Transpose[x]]),
          Message[OPInv::matrixQ, x],
        FreeQ[OperatorQ /@ Flatten[x], True],
          Inverse[x],
        StrictUpperDiagonalMatrixQ[x], 
          TimeConstrained[upperinv[x], (TimeConstraint/.{opts}/.Options[OPInv])],
        StrictLowerDiagonalMatrixQ[x], 
          TimeConstrained[lowerinv[x], (TimeConstraint/.{opts}/.Options[OPInv])],
        DiagonalMatrixQ[x],            
          DiagonalMatrix[Table[OPExpand[Power[x[[i,i]],-1]],{i,1,Length[x]}]],
        True, 
          Module[{tmp=GaussDecompose[x], invu,invd,invl},
            invu=OPInv[tmp[[3]]]; invd=OPInv[tmp[[2]]]; invl=OPInv[tmp[[1]]];
            refreshDiamond[];
            Return[
              TimeConstrained[
                 invu \[Diamond] invd \[Diamond] invl,
                 (TimeConstraint/.{opts}/.Options[OPInv])
              ]
            ]
          ]
  ];

Options[OPInv] = {TimeConstraint -> 300}

OPInv::matrixQ=
"`1` is not a square matrix and so cannot calculate the inverse. "

(* ----- Codes below are aimed at a Gauss decomposition routine ----- *)

(* Noncommutative version of LUFactor, modifies the LUFactor routine in the package
LinearAlgebra`GaussianElimination` by removing the pivot and changing Times into Diamond *)

LUFacOPMatrix[aa_] := 
  Module[{a = aa, ii, i, j, mpiv, m, n = Length[aa]}, 
    For[ii = 1, ii <= n, ii++,
      mpiv = a[[ii, ii]];
      For[i = ii + 1, i <= n, i++, 
        m = a[[i, ii]] = Diamond[a[[i, ii]], mpiv^(-1)];
        For[j = ii + 1, j <= n, j++, a[[i, j]] -= m \[Diamond] a[[ii, j]]]
      ];
    ];
    a
  ];


GaussD[aa_?MatrixQ] := DiagonalPart[ LUFacOPMatrix[aa] ];

(*
GaussL[aa_?MatrixQ] := LUMatrices[ LUFacOPMatrix[aa] ][[1]];
GaussU[aa_?MatrixQ] := inv[GaussD[aa]] \[Diamond] 
    LUMatrices[LUFacOPMatrix[aa] ][[2]];
*)

(* The codes below avoids calculating LUFacOPMatrix[aa] twice, so faster *)

GaussL[aa_?MatrixQ] := 
    Module[{tmp,ret,ans,l=Length[aa]},
       tmp = LUFacOPMatrix[aa];
       For[i=1,i<=l,i++,
         For[j=1,j<i,j++,
           ret[i,j]=tmp[[i,j]]
         ]
       ];
       ans=StrictLowerDiagonalMatrix[ret,l];
       Return[ans]
    ];

GaussU[aa_?MatrixQ] := 
    Module[{tmp,ret,l=Length[aa],ans},
      tmp = LUFacOPMatrix[aa];
      For[i=1,i<=l,i++,
        For[j=i+1,j<=l,j++,
           ret[i,j]=tmp[[i,i]]^(-1) \[Diamond] tmp[[i,j]]
        ]
      ];
      ans=StrictUpperDiagonalMatrix[ret, l];
      Return[ans]
    ];

GaussDecompose[aa_?MatrixQ] := 
    Module[{tmp,ans,ret,l=Length[aa]},
      tmp=LUFacOPMatrix[aa];
      For[i=1,i<=l,i++,
         For[j=1,j<i,j++,
           ret[i,j]=tmp[[i,j]]
         ];
         For[j=i+1,j<=l,j++,
           ret[i,j]=tmp[[i,i]]^(-1) \[Diamond] tmp[[i,j]]
         ]
      ];
      ans[1]=StrictLowerDiagonalMatrix[ret,l];
      ans[3]=StrictUpperDiagonalMatrix[ret,l];
      ans[2]=DiagonalMatrix[Table[tmp[[i,i]],{i,1,l}]];
      Return[{ans[1],ans[2],ans[3]}]
    ];              

(* Solving linear equations consisted of operators *)

OPLinearSolve[a_?ListQ, b_?ListQ] :=
    If[Not[MatrixQ[b]] && Not[MatrixQ[a]],
       Message[OPLinearSolve::dfn],
       Which[MatrixQ[a] && Length[a] == Length[Transpose[a]] && 
               Length[Dimensions[b]] == 1 && Dimensions[b][[1]]==Length[a],
             Module[{tmp=Transpose[{b}],gaussdec,ret},
             (*
                 gaussdec=GaussDecompose[a];
                 ret= inv[gaussdec[[3]]] \[Diamond] inv[gaussdec[[2]]] \[Diamond] 
                      inv[gaussdec[[1]]] \[Diamond] tmp;*)
                 ret=OPInv[a] \[Diamond] tmp;
                 Transpose[ret][[1]]
             ],
             MatrixQ[b] && Length[b] == Length[Transpose[b]] && 
               Length[Dimensions[a]] == 1 && Dimensions[a][[1]]==Length[a],
             Block[{tmp={a},gaussdec,ret},
             (*    
                 gaussdec=GaussDecompose[b];
                 ret = tmp \[Diamond] inv[gaussdec[[3]]] \[Diamond] 
                       inv[gaussdec[[2]]] \[Diamond] inv[gaussdec[[1]]];*)
                 ret = tmp \[Diamond] OPInv[b];
                 ret[[1]]
             ],
             True, Message[OPLinearSolve::dfn]
       ]
    ];

OPLinearSolve::dfn=
"The linear equation is ill-defined. Check the coefficient matrix etc."

NCPowerToDiamond[a_NCPower]:=
  Which[Not[MatrixQ[a[[1]]]] || 
           Length[a[[1]]]=!=Length[Transpose[a[[1]]]] || Not[IntegerQ[a[[2]]]],
        Message[NCPowerToDiamond::err],
        Positive[a[[2]]],
            Diamond@@Table[a[[1]],{a[[2]]}]//OPExpand,
        Negative[a[[2]]],
            Diamond@@Table[NCPower[a[[1]],-1],{-a[[2]]}]
  ];

NCPowerToDiamond::err=
"Something is wrong in the arguments: either the matrix is \
not square or the power is not integer."

(* Operator transpose *)

Tp[Diamond[x_, y___]] := Diamond[Tp[Diamond[y]], Tp[x]];
Tp[x_ + y_] := Plus[Tp[x], Tp[y]];
Tp[Literal[Times[x_, y_]]] := Times[Tp[y], Tp[x]];
Tp[x_] := x /; Not[OperatorQ[x]] || Length[x] <= 1;
Tp[x_] := 
    Table[Tp[x[[j, i]]], {i, 1, Length[x]}, {j, 1, Length[Transpose[x]]}] /; 
      MatrixQ[x];
Tp[- x_] := -Tp[x];
Tp[Power[x_, y_]] := Power[Tp[x], Tp[y]];



(* ------------------------------------------------------ *)
Select[{DProd, CircleTimes, CirclePlus, CircleCommutator,
      RevDiamond,ComplementMatrix,LeftMinor,RightMinor,
      LeftDet,RightDet,ncinv,OPInv,DiagonalPart,
      LUFacOPMatrix, GaussL, GaussD, GaussU, GaussDecompose, 
      OPLinearSolve,NCPowerToDiamond,Tp,
      StrictUpperDiagonalMatrix,StrictLowerDiagonalMatrix},
      SetAttributes[#,ReadProtected]&];
(* ------------------------------------------------------ *)

End[]

Protect[CircleTimes, DProd, Dot, Times, Diamond, Commutator, 
        NCPower, OPExpand, CirclePlus, CircleCommutator,
        RevDiamond,ComplementMatrix,LeftMinor,RightMinor,
        LeftDet,RightDet,ncinv,OPInv,DiagonalPart,
        LUFacOPMatrix, GaussL, GaussD, GaussU, GaussDecompose, 
        OPLinearSolve,NCPowerToDiamond,Tp,
        StrictUpperDiagonalMatrix,StrictLowerDiagonalMatrix];

EndPackage[]
