(*: Mathematica Version: 3.0, Not tested for earlier versions. *)

(*: The OPScalar Package *)

(* Author:  Liu Zhao 
            Inst. Mod. Phys., Northwest Univ., Xian 710069, China
            lzhao@phy.nwu.edu.cn
            
            Any comments, bug reports, suggestions etc welcomed.
*)

(*: Context: OPLinearAlgebra`OPScalar`  *)

(*: Summary: A general purpose noncommutative multiplication and commutator
evaluation package.
*)

(*: Keywords: Commutator, noncommutative multiplication
*)

(*: Requirements: None.  *)

(*: Package Version: 1.1 *)

(*: History: Created in May, 2000 by Liu Zhao 
             Minor bug fixes in April, 2002 
	     Modified definition of Commutator so that it now can be used to
	     evaluate Lie algebra products (April, 2003)
*)


(*: Source: the definition of the basic functionalities for Diamond are borrowed
    from the definition of NonCommutativeMultiply in the package NCAlgebra 
    (uses 5 lines of NCAlgebra package codes
*)


BeginPackage["OPLinearAlgebra`OPScalar`"];

(*
Initializing the environment so that no memory confliction would occur if
the package is loaded more than once.
*)

Unprotect[Minus, Times, Power, Expand];
Unprotect[SetOperators, ClearOperators, OperatorQ, SetCommutator, Diamond,
          Commutator, CommutatorQ, ClearCommutator, CommutatorToDiamond,
          ClearAllOperators, CanonicalOrder,
          ListBasicOperators, SetDiamond, ClearDiamond,
          OPExpand, NCPower, SetAllOperators, SetAllOperatorsQ,
          PermitTimes, CollectPower,PowerToDiamond,DiamondToPower,CM];

Clear[Times, SetOperators, ClearOperators, OperatorQ, SetCommutator, Diamond,
      Commutator, CommutatorQ, ClearCommutator, CommutatorToDiamond,
      ClearAllOperators, CanonicalOrder,
      ListBasicOperators, SetDiamond, ClearDiamond,
      OPExpand, NCPower, SetAllOperators, SetAllOperatorsQ,
      PermitTimes, CollectPower,PowerToDiamond,DiamondToPower,CM];

(* Brief helps for new functions *)

ClearCommutator::usage=
"ClearCommutator[x, y] clears the value of [x, y], if it has been assigned \
by SetCommutator[x, y, z]."

SetOperators::usage=
"SetOperators[x,y,z etc] declears its arguments to be operators."

OperatorQ::Usage=
"OperatorQ[x] returns the value True if the expression x is declared \
as operator, and False otherwise. Used to define the rules for Commutator."

SetAllOperators::usage=
"Setting SetAllOperators[] declares everything except numerical variables \
to be operators. This is not true by default. Setting this to be true \
will speed up the computation of noncommutative multiplications. \
Alternatively, users can declare their own operators by use of SetOperators, \
and once done so, only user declared objects and their compositions are \
regarded as operators."

SetAllOperatorsQ::usage=
"SetAllOperatorsQ[] determines whether SetAllOperators is being activated."

ClearOperators::usage=
"ClearOperators[x, y, z] removes its arguments from the list of user declared \
operators."

SetCommutator::usage=
"SetCommutator[x,y,z] assigns the Commutator of x and y to be z, i.e. \
Commutator[x, y]= -Commutator[y, x] = z."

ClearAllOperators::usage=
"ClearAllOperators[] removes all user defined operators except matrices. \
Once invoked, the effect of SetAllOperators is also cleared."

ListBasicOperators::usage=
"ListBasicOperators[] gives the list of all basic user-defined operators."

Diamond::usage=
"Diamond[a,b,c ...] gives the noncommutative multiplication for operators. \
This version of noncommutative multiplication is associative, and its \
evaluation precedence is higher than Times, i.e. a \[Diamond] b c \[Diamond] d is
considered as (a \[Diamond] b)(c \[Diamond] d), rather \
than a \[Diamond] (b c) \[Diamond] d, but lower than Power, i.e. \
a^c \[Diamond] b is interpreted as (a^c) \[Diamond] d. Caution has to be taken
when using complicated expressions containing \[Diamond]."

Commutator::usage=
"Commutator[a,b] evaluates the commutator of the two operator expressions \
a and b. It recognizes all canonical rules for commutators, especially \
the Lebnitz rule and the Jacobi Identity.

Important: One cannot assign values to Commutator[x,y] directly using \
the commands Commutator[x,y]:=sth or Commutator[x,y]=sth, because \
the definition of Commutator[x, y] is protected. Use SetCommutator instead. \
Likewise, to remove the value of [x, y] assigned by SetCommutator[x, y, z], \
use ClearCommutator[x,y]."

CommutatorQ::usage=
"CommutatorQ[x] returns True if Head[x] is Commutator and False otherwise."


CommutatorToDiamond::usage=
"ConvertToCommutator[x] turns a commutator into antisymmetric diamond products."


OPExpand::usage =
"OPExpand[expr] expands out \
Diamond's in expr. For example, OPExpand[a \[Diamond] (b+c)] \
will result in a \[Diamond] b + a \[Diamond] c.";

SetDiamond::usage =
"SetDiamond[a, b, c] assigns the value c to a \[Diamond] b, \
and simutanuously transfer the rule into OPExpand to \
help simplification."

ClearDiamond::usage =
"ClearDiamond[x, y] clears the assigned value for x \[Diamond] y \
as well as the corresponding simplification rules."

NCPower::usage=
"NCPower[x, n] is equivalent to Power[x, n] if x is not a matrix. It gives a \
noncommutative generalization of Power if x is a matrix (the system function
Power operates incorrectly for matrix bases)."

(*
setDiamondRules::usage =
"setDiamondRules recordes rules for value assignment to Diamond[a, b]. \
It is recommended that the user frequently check the value of \
setDiamondRules so that the proper rules are used in the calculation."
*)

CollectPower::usage=
"On[CollectPower] enables  powercollection in noncommutative multiplications. \
Off disables this feature. Default is on. PowerToDiamond turns this feature Off."

PermitTimes::usage=
"On[PermitOperatorTimes] enables the definition of usual Times operation between \
operators. Off[PermitOperatorTimes] disables this feature. Default is off."

PowerToDiamond::usage=
"PowerToDiamond[expr] converts all positive integer powers in expr into successive \
Diamond products of the bases, and all negative integer powers in expr into \
successive Diamond products of the inverse of the bases. It has no effect if expr \
does not contain any power expression or all the exponents are nonnumeric. \
This function has no effect if CollectPower is On."

CanonicalOrder::usage=
"CanonicalOrder[expr] gives a canonical ordering foe Diamond expressions. Tips: \
this function can be used to define Normal Ordering of Heisenberg operators \
provided the naming of Heisenberg operators follows the standard convention \
(i.e. the lowering operators precede the raising operators in lexicographical order)."

DiamondToPower::usage=
"DiamondToPower[expr] does the inverse of PowerToDiamond, i.e. it changes all 
expressions like x \[Diamond] x into x^2 etc. This function is effective in any modes, \
regardless of whether CollectPower is On or Off."

CM::usage=
"CM is the Output form of Commutator. If called explicitly, it will give \
essentially the same result as Commutator, but without expansionoveer \
Diamond expressions."


Begin["`Private`"]

(* Initialization of the variable Operators *)

(* "loaded" introduced to keep declared operators in case the package is reloaded *)
(* 29 June, 2000 *)
 
If[loaded=!=True,loaded:=False];
If[Not[loaded],Operators := {CM, Commutator, Diamond}];
loaded:=True;

AllOperators := False;
powercollect := True;

SetOperators[x__] :=
    Module[{userOperators,tmp}, AllOperators = False; tmp=Flatten[{x}];
       Which[Select[tmp,NumericQ]=!={}, Message[SetOperators::num],
             Select[tmp,simplesymbolQ]=!=tmp, Message[SetOperators::symbol],
             True,
               Block[{},
                 Operators = Union[Operators, Flatten[tmp]];
                 userOperators = Complement[Operators,{Commutator, CM, Diamond}];
                 Return[userOperators]
               ]
       ]
    ];

SetOperators::num=
"Attempting to set numeric objects as operators! Canceled."

SetOperators::symbol=
"Only atomic symbols or atomic symbols with a subscript can be declared \
as basic operators."


ClearOperators[x__] :=
    Block[{}, AllOperators=False;
       Operators = Complement[Operators, Flatten[{x}]]; 
       Return[{x}]
    ];

simplesymbolQ[x_]:=
    Which[AtomQ[x], True,
          Head[x]===Subscript && AtomQ[x[[1]]], True,
          True, False
    ];


SetAllOperators /: On[SetAllOperators] :=
    Block[{},
      AllOperators = True; 
      Message[SetAllOperators::ison]
    ];

SetAllOperators /: Off[SetAllOperators] :=
    Block[{},
      AllOperators = False;
      Message[SetAllOperators::isoff]
    ];

SetAllOperatorsQ[]:= AllOperators;

SetAllOperators::ison=
"Everything except numeric variables are now operators."

SetAllOperators::isoff=
"Now only user defined operators are in effect. Use SetOperators to declare \
more user operators."


SetAllOperators[]:=
    Block[{}, AllOperators = True;
       Message[SetAllOperators::ison]
    ];

ClearAllOperators[] :=
    If[Not[AllOperators],
       Block[{userOperators},
          userOperators = Complement[Operators,{Commutator, CM, Diamond}];
          Operators={Commutator, CM, Diamond};
          AllOperators=False;
          Return[SequenceForm["All user defined operators, including ",
                               userOperators,", removed."
                 ]
          ]
       ],
       AllOperators = False; Operators={Commutator, CM, Diamond};
       Message[SetAllOperators::isoff]
    ];


ListBasicOperators[]:=
    If[AllOperators, Print[SetAllOperators::ison],
       Block[{userOperators},
          userOperators = Complement[Operators,{Commutator, CM, Diamond}];
          Return[DeleteCases[userOperators,_?NumericQ]]
       ]
    ];


(* Determining whether an expression is an operator *)

OperatorQ[x_] :=
    If[AllOperators, Not[NumericQ[x]],
        Which[(MemberQ[Operators,x] && Not[NumericQ[x]]) ||
                       MemberQ[Operators,Head[x]] || MatrixQ[x], True,
              NumericQ[x], False,
              True,    If[Length[x] == 0, False,
                       OperatorQ[Head[x]] ||
                       OperatorQ[Part[x,1]] || OperatorQ[Rest[x/.Head[x]->1]]
                       ]
        ]
    ];

squareMatrixQ[x_]:=
    MatrixQ[x] && Length[x] == Length[Transpose[x]];


(* Defining the noncommutative product operation Diamond[] and
a noncommutative version of Power[], pretty formatting *)

(* NCMultiplication Codes *)

Literal[Diamond[a___, Diamond[b__], c___]] :=
    Diamond[a, b, c];

Literal[Diamond[a___, b_ c_, d___]]:=
    b Diamond[a, c, d] /; Not[OperatorQ[b]];

Literal[Diamond[a___, b_, c___]] :=
    b Diamond[a, c] /; Not[OperatorQ[b]];

Literal[Diamond[a_]] := a;

Diamond[] := 1;

(* My own codes *)

(* Codes which do not depend on CollectPower mode *)

Literal[Diamond[x___, Power[Diamond[y_, w___], -1], z___]] :=
  Diamond[x, (Diamond[w])^-1, y^-1, z];

Literal[Diamond[x___, NCPower[Diamond[y_, w___], -1], z___]] :=
  Diamond[x, NCPower[Diamond[w],-1], NCPower[y,-1], z];

Literal[Diamond[x___, y_Plus, z___]] :=
  - Diamond[x, abs[y], z] /; abs[y]=!=y;


(* Codes which depend on CollectPower mode *)

defineDiamond[]:=
Module[{},
Diamond[a___, b_, b_, c___] := Diamond[a, NCPower[b,2], c] /;
    powercollect && Not[MatrixQ[b]];
Literal[Diamond[a___,  NCPower[b_,  n_], b_, c___]] :=
    Diamond[a, NCPower[b,n + 1], c] /; powercollect &&
          (Not[OperatorQ[n]] || Not[OperatorQ[b]]);
Literal[Diamond[a___, b_, NCPower[b_, n_], c___]] :=
    Diamond[a, NCPower[b,n + 1], c] /; powercollect &&
          (Not[OperatorQ[n]] || Not[OperatorQ[b]]);
Literal[Diamond[a___,  NCPower[b_,  m_], NCPower[b_, n_], c___]] :=
    Diamond[a, NCPower[b,m + n], c] /; powercollect &&
          (Commutator[m,n]==0 && Commutator[b,n]==0 && Commutator[b,m]==0);
Diamond[a___,  Power[b_,  n_], b_, c___] :=
    Diamond[a, Power[b,n + 1], c] /; powercollect &&
          (Not[OperatorQ[n]] || Not[OperatorQ[b]]);
Diamond[a___, b_, Power[b_, n_], c___] :=
    Diamond[a, Power[b,n + 1], c] /; powercollect &&
          (Not[OperatorQ[n]] || Not[OperatorQ[b]]);
Diamond[a___,  Power[b_,  m_], Power[b_, n_], c___] :=
    Diamond[a, Power[b,m + n], c] /; powercollect &&
         (Commutator[m,n]==0 && Commutator[b,n]==0 && Commutator[b,m]==0);
(* This concerns the output of Diamond expressions in CollectPower mode *)
Diamond /: MakeBoxes[Diamond[x_, y___], TraditionalForm] :=
  RowBox[{RowBox[{If[Head[x]===Plus, 
                     Seq["(",MakeBoxes[x,TraditionalForm],")"], 
                     MakeBoxes[x,TraditionalForm]
                  ]/.Seq->Sequence}], 
          If[Length[{y}] === 0, "",
             Seq["\[Diamond]",MakeBoxes[Diamond[y], TraditionalForm]]
          ]/.Seq->Sequence}
  ]/;powercollect;
Diamond /: MakeBoxes[Diamond[x_, y___], StandardForm] :=
  RowBox[{RowBox[{If[Head[x]===Plus, 
                     Seq["(",MakeBoxes[x,StandardForm],")"], 
                     MakeBoxes[x,StandardForm]
                  ]/.Seq->Sequence}], 
          If[Length[{y}] === 0, "",
             Seq["\[Diamond]",MakeBoxes[Diamond[y], StandardForm]]
          ]/.Seq->Sequence}
  ]/;powercollect;
SC[y___, x_, x_, z___] := SC[y, x^2, z] /; 
  powercollect && Not[MemberQ[{Plus,Diamond},Head[x]]];
SC[y___, x_, x_^n_, z___] := SC[y, x^(n + 1), z] /; 
  powercollect && Not[MemberQ[{Plus,Diamond},Head[x]]];
SC[y___, x_^n_, x_, z___] := SC[y, x^(n + 1), z] /; 
  powercollect && Not[MemberQ[{Plus,Diamond},Head[x]]];
]; (* end defineDiamond *)

defineDiamond[]; (* run defineDiamond[] once to define the function Diamond *)

(* SC is a light version of Diamond, useful for expanding complicated expressions *)

Literal[SC[y___, a_ x_, z___]] := a SC[y, x, z] /; Not[OperatorQ[a]];

(* abs function: called from within Diamond *)

abs[x_] := x /; Head[x] =!= Plus;
abs[x_Plus] := 
  Module[{ss}, ss = x[[1]]; 
    If[Head[ss] === Times && NumericQ[ss[[1]]] && ss[[1]] < 0, Return[-x], x]
  ];
SetAttributes[abs,Listable];

(* Power and NCPower definitions *)

Power[Diamond[y_, w___], -1] := Diamond[Power[Diamond[w],-1], y^-1];
Power[x_, n_] := 
    Power[Power[x,-1],-n] /; Head[x]===Diamond && NumericQ[n] && 
       Negative[n] && n=!=-1;
Power[x_Plus, n_] := (-1)^n Power[abs[x], n] /; abs[x]=!=x && NumericQ[n];

NCPower[x_, 1] := x;
NCPower[x_, n_] := Power[x, n]/;
    FreeQ[x, NCPower] && Not[MatrixQ[x]] && FreeQ[x, _?MatrixQ];

Literal[NCPower[NCPower[b_,  n_], m_]] :=
    NCPower[b, n m] /; Not[OperatorQ[n]] && Not[OperatorQ[m]] (*&& 
      Abs[n]=!=1 && Abs[m] =!=1*);

NCPower[x_, 0] :=
    Table[If[i==j,1,0],{i,1,Length[x]},{j,1,Length[x]}] /;MatrixQ[x];

NCPower[Diamond[y_, w___], -1] := 
    Diamond[NCPower[Diamond[w],-1], NCPower[y,-1]];
NCPower[x_, n_] := 
    NCPower[NCPower[x,-1],-n] /; Head[x]===Diamond && NumericQ[n] && 
       Negative[n] && n=!=-1;


(* CollectPower modes switches *)

CollectPower[] := powercollect;

CollectPower /: On[CollectPower] :=
    Block[{},powercollect=True;
          Print["CollectPower is On."]
    ];

CollectPower /: Off[CollectPower] :=
    Block[{},powercollect=False;
          Print["CollectPower is Off."]
    ];

NCPower/: MakeBoxes[NCPower[x_, y_], TraditionalForm] :=
  RowBox[{SuperscriptBox[If[MemberQ[{Plus, Times, Diamond, CircleTimes}, Head[x]],
                             RowBox[{"(",MakeBoxes[x,TraditionalForm],")"}], 
                             MakeBoxes[x,TraditionalForm]
                         ], MakeBoxes[y,TraditionalForm]]}
  ];

NCPower/: MakeBoxes[NCPower[x_, y_], StandardForm] :=
  RowBox[{SuperscriptBox[If[MemberQ[{Plus, Times, Diamond, CircleTimes}, Head[x]],
                             RowBox[{"(",MakeBoxes[x,StandardForm],")"}], 
                             MakeBoxes[x,StandardForm]
                         ], MakeBoxes[y,StandardForm]]}
  ];

Power/: MakeBoxes[Power[x_, y_], TraditionalForm] :=
  RowBox[{SuperscriptBox[If[MemberQ[{Plus, Times, Diamond, CircleTimes, Power}, Head[x]],
                             RowBox[{"(",MakeBoxes[x,TraditionalForm],")"}], 
                             MakeBoxes[x,TraditionalForm]
                         ],  MakeBoxes[y,TraditionalForm]]}
  ];

Power/: MakeBoxes[Power[x_, y_], StandardForm] :=
  RowBox[{SuperscriptBox[If[MemberQ[{Plus, Times, Diamond, CircleTimes, Power}, Head[x]],
                             RowBox[{"(",MakeBoxes[x,StandardForm],")"}], 
                             MakeBoxes[x,StandardForm]
                         ],  MakeBoxes[y,StandardForm]]}
  ];



(* Expand noncommutative products into smallest factors *)
(*
ExpandDiamondPower[Power[x_, n_]] := ExpandDiamondPower[Power[x,n]] =
    Which[Not[NumericQ[n]], Power[x, n],
          Not[OperatorQ[x]], Expand[Power[x,n]],
          Positive[n],
            Which[Not[IntegerQ[n]], Power[x, n],
                  n===2,
                  Block[{var,ret},SetOperators[var];
                     ret=OPExpand[Diamond@@{x,var,x}]/.var->1;
                     ClearOperators[var];
                     Return[ret]
                  ],
                  True,
                  OPExpand[Diamond[x, ExpandDiamondPower[x^(n - 1)]]]
            ],
          Negative[n],
            If[Head[x]===Diamond,
              OPExpand[Power[Power[x,-1],-n]],
              Power[x,n]
            ]
    ];
*)
ExpandDiamondPower[Power[x_, n_]] :=(* ExpandDiamondPower[Power[x,n]] = *)
    Which[Not[NumericQ[n]], Power[x, n],
          Not[OperatorQ[x]], Expand[Power[x,n]],
          Positive[n],
            Which[Not[IntegerQ[n]], Power[x, n],
                  True,
                  ExpandSC[x, n] 
            ],
          Negative[n],
            If[Head[x]===Diamond,
              OPExpand[Power[Power[x,-1],-n]],
              Power[x,n]
            ]
    ];

ExpandSC[x_, n_] :=
  SC@@Table[x,{n}]//.
    {Literal[SC[a___, b_Plus, c___]] :> (SC[a, #, c] & /@ b)} /.SC->Diamond;
       
(*
OPExpand[Power[a_, n_]] :=
    If[Length[a]<=1, If[IntegerQ[n], PowerToDiamond[Power[a,n]], Power[a, n]],
      ExpandDiamondPower[Power[a,n]]
    ];

diamondexp[Times[x_, y___]]:=
   Expand[OPExpand[x] ExpandDiamond[Times[y]]];
diamondexp[Diamond[x_, y___]]:=
   Expand[OPExpand[x] \[Diamond] OPExpand[Diamond[y]]];
diamondexp[x_ + y_]:=
   Expand[OPExpand[x] + OPExpand[y]];
diamondexp[x_] := x /; Not[MemberQ[{Times,Plus,Diamond},Head[x]]];
*)

OPExpand[expr_] :=
  Module[{rulelist,fun},
    rulelist=
      Union[{ Power[x_Plus,n_Integer?Positive]:>ExpandDiamondPower[Power[x,n]],
              Power[x_Diamond,n_Integer?Positive]:>ExpandDiamondPower[Power[x,n]],
              Literal[Diamond[a___, b_Plus, c___]] :> (Diamond[a, #, c]& /@ b)
            }, setDiamondRules
      ];
    fun=#//.rulelist&;
    Expand[MapAll[fun, expr]]
  ];

(* Setting rules of Diamond value substitutions *)

setDiamondRules={};

SetDiamond[x_, y_, z_] :=
    Module[{},Unprotect[Diamond];
      setDiamondRules =
        Union[setDiamondRules,{ Literal[Diamond[a___, x, y, b___]] :>
                                   Diamond[a, z, b],
                                Literal[Diamond[a___, x^n_, y, b___]] :>
                                   Diamond[a, x^(n - 1), z, b],
                                Literal[Diamond[a___, x, y^n_, b___]] :>
                                   Diamond[a, z, y^(n - 1), b]
                               }
        ];
      Diamond[x, y] := z;
      Protect[Diamond];
      Return[z]
    ];

ClearDiamond[x_, y_] :=
    Module[{},Unprotect[Diamond];
       Diamond[x, y]=.;
       setDiamondRules = Union[Select[setDiamondRules, FreeQ[#, x]&],
                               Select[setDiamondRules, FreeQ[#, y]&]
                         ];
       Protect[Diamond];
       Print["Value and rules for ", x, " \[Diamond] ", y, " cleared."]
    ];

(* Definition of Commutator and CommutatorQ *)

CommutatorQ[x_] := If[Head[x] === CM, True, False];

CM[x_ + y_, z_] := CM[x, z] + CM[y, z];
CM[x_, y_ + z_] := CM[x, y] + CM[x, z];
CM[x_, y_] := 0 /; (x == y) || (Not[OperatorQ[x]]) || (Not[OperatorQ[y]]);

CM[y_, x_] := -CM[x, y] /;
    And[ordQ[{x, y}], Not[CommutatorQ[x]], Not[CommutatorQ[y]]];

CM[CM[a_, b_], c_]:= -CM[c, CM[a, b]] /;
    And[Not[CommutatorQ[c]], CommutatorQ[CM[a, b]]
    ];

(*
CM[a_, CM[b_, c_]] :=
  - CM[c, CM[a, b]] + CM[b, CM[a, c]] /;
      And[ordQ[{b,a}], ordQ[{c,a}], CommutatorQ[CM[b, a]],
             CommutatorQ[CM[c, a]]
      ];
*)   

(* The above is commented out and the next definition added on April 2, 2003 *)

CM[a_, CM[b_, c_]] :=
  - CM[c, CM[a, b]] + CM[b, CM[a, c]] /;
      And[Not[CommutatorQ[CM[a, b]]],
          Not[CommutatorQ[CM[a, c]]],
	  Or[Not[CommutatorQ[c]], Not[CommutatorQ[b]]]
      ];

CM[CM[a_, b_], CM[c_, d_]]:= - CM[b, CM[a, CM[c, d]]] + CM[a, CM[b, CM[c, d]]]; 

CM[Minus[a_], b_] := - CM[a, b];
CM[a_, Minus[b_]] := - CM[a, b];

CM[x_*y___, z_] :=
    x \[Diamond] CM[Times[y], z] + CM[x, z] \[Diamond] Times[y];
CM[x_, y_*z___] :=
    y \[Diamond] CM[x, Times[z]] + CM[x, y] \[Diamond] Times[z];
CM[x_, Diamond[y_,z___]] :=
    y \[Diamond] CM[x, Diamond[z]] + CM[x, y] \[Diamond] Diamond[z];
CM[Diamond[x_,y___], z_] :=
    x \[Diamond] CM[Diamond[y], z] + CM[x, z] \[Diamond] Diamond[y];

CM[x_^n_, y_] :=
    x\[Diamond]CM[x^(n - 1), y] +
       CM[x, y]\[Diamond] x^(n - 1) /;
           (IntegerQ[n] && Positive[n]);
CM[x_, y_^n_] :=
    y\[Diamond]CM[x, y^(n - 1)] +
       CM[x, y]\[Diamond] y^(n - 1) /;
           (IntegerQ[n] && Positive[n]);

CM[fun_[x_],y_] := 0 /;
  (CM[x,y]==0 && Not[OperatorQ[fun]]);

CM[y_,fun_[x_]] := 0 /;
  (CM[x,y]==0 && Not[OperatorQ[fun]]);

Commutator[x_,y_]:= CM[x,y] //OPExpand;

(* Manipulating values of Commutator *)

SetCommutator[x_, y_, z_] := Module[{},Unprotect[CM];
    If[ordQ[{x, y}], CM[x, y] := z, CM[y, x] := -z];
    Protect[CM];];

ClearCommutator[x_, y_] := Module[{},Unprotect[CM];
    If[ordQ[{x, y}], CM[x, y]=., CM[y, x]=.];
    Protect[CM];];


(* Formatting output for Commutator and Diamond *)

(*
Commutator/: MakeBoxes[Commutator[x_,y_],TraditionalForm]:=
  RowBox[{"[",
          MakeBoxes[x,TraditionalForm],",",MakeBoxes[y,TraditionalForm],
          "]"
  }];
*)
CM/: MakeBoxes[CM[x_,y_],TraditionalForm]:=
  RowBox[{"[",
          MakeBoxes[x,TraditionalForm],",",MakeBoxes[y,TraditionalForm],
          "]"
  }];
CM/: MakeBoxes[CM[x_,y_],StandardForm]:=
  RowBox[{"[",
          MakeBoxes[x,StandardForm],",",MakeBoxes[y,StandardForm],
          "]"
  }];

(* strict ordering function, used in definition of Commutator *)

ordQ[{x___}] := ordQ[x];
ordQ[x___] := (Signature[Flatten[{x}]] > 0) && OrderedQ[Flatten[{x}]] /; 
    FreeQ[{x}, CM];
ordQ[x___, y_] := True /; FreeQ[{x}, CM] && CommutatorQ[y];
ordQ[x___CM] := OrderedQ[{x}];
ordQ[x___, -y_CM] := True /; FreeQ[{x}, CM];

(*
ordQ[x___, Commutator[y_, z_], w___] := ordQ[x, y, z, w];
ordQ[x___] := 
    OrderedQ[Flatten[{x}]] /; FreeQ[{x}, Commutator] && FreeQ[{x}, CM];
ordQ[x___, CM[y_, z_], w___] := ordQ[x, y, z, w];
ordQ[{x___}]:=ordQ[x];
*)


(* Conversion between Commutator and noncommutative multiplications *)

CommutatorToDiamond[x_] :=
    x//.{CM[a_, b_]:> Diamond[a, b] - Diamond[b, a]} // OPExpand  ;

SetAttributes[CommutatorToDiamond,{Listable}];

(* Conversion between Power and Diamond *)
(* ---- Important: PowerToDiamond has no effect if CollectPower is On. ---- *)

SetAttributes[DiamondToPower,Listable];
SetAttributes[PowerToDiamond,Listable];


PowerToDiamond[expr_] := 
  Module[{piQ, ne2Q,tmp=powercollect,ret}, 
    powercollect=False; 
    Unprotect[Diamond]; 
    Clear[defineDiamond];
    defineDiamond[];
    piQ[x_] := IntegerQ[x] && Positive[x];
    ne2Q[x_] := IntegerQ[x] && Negative[x] && (x <= -2);
    ret=expr //. {Literal[Power[x_, y_?piQ]] :> Apply[Diamond, Table[x, {y}]],
              Literal[Power[x_, y_?ne2Q]]:> Apply[Diamond, Table[x^-1, {-y}]]
             };
    powercollect=tmp;
    Protect[Diamond];
    Return[ret]
  ];

DiamondToPower[expr_] := 
  Module[{tmp=powercollect,ret}, 
    powercollect=True; 
    Unprotect[Diamond]; 
    Clear[defineDiamond];
    defineDiamond[];
    ret=expr;
    powercollect=tmp;
    Protect[Diamond];
    Return[ret]
  ];

(* Canonical Ordering of Diamond expressions *)

op[z___, x_, y_, w___] := 
  op[z, y, x, w] + op[z, Commutator[x, y], w] /; Not[OrderedQ[{x, y}]];

Literal[op[z___, x_ y_, w___]] := x op[z, y, w] /; Not[OperatorQ[x]];

Literal[op[z___, x_, w___]] := x op[z, w] /; Not[OperatorQ[x]];

Literal[op[z___, op[x___], w___]] := op[z, x, w]; 

op[] := 1;

op[x___, 0, y___] := 0;
op[x___, y_, z___] := y op[x, z] /; Not[OperatorQ[y]];

op[x_ + y_] := op[x] + op[y];

op[z___,x_Diamond,y___] := op[z, x//. Diamond -> Sequence, y];


CanonicalOrder[x_] := op[x] /. {op->Diamond};

(* Something to care about the correct operation of calculus and matrix operations *)

Unprotect[D,Integrate,Inverse];

D[Diamond[x_, y___], z_] :=
  D[x, z]\[Diamond]Diamond[y] + x\[Diamond]D[Diamond[y], z];

D[Plus[a___,Diamond[x_, y___],b___], z_] :=
  D[Plus[a,b],z] + D[x, z]\[Diamond]Diamond[y] + x\[Diamond]D[Diamond[y], z];

Integrate[x_?OperatorQ, y_] := Message[Integrate::op];
Integrate[x_, y_?OperatorQ] := Message[Integrate::op];

Integrate::op=
"Currently there is no integration rutines for operator-valued expressions."

Inverse[x_?MatrixQ]:=
   OPLinearAlgebra`OPMatrix`OPInv[x] /; Not[FreeQ[OperatorQ /@ Flatten[x], True]];

Protect[D,Integrate,Inverse];

(* Pretty printing for negative powers *)

PrettyTimes[]:=Module[{},
Unprotect[Times];
Times /: MakeBoxes[Times[x_, y___], TraditionalForm] :=
   RowBox[{RowBox[{If[Head[x]===Plus || Head[x]===Complex, 
                    Seq["(",MakeBoxes[x,TraditionalForm],")"], 
                    If[x===-1, "- ",MakeBoxes[x,TraditionalForm]]
                   ]/.Seq->Sequence
         }], 
         Which[Length[{y}] === 0, "",
               Length[{y}] === 1,
                 If[Head[y]===Plus, 
                     Seq["(",MakeBoxes[y, TraditionalForm],")"],
                     MakeBoxes[y, TraditionalForm]
                 ]/.Seq->Sequence,
               True, MakeBoxes[Times[y], TraditionalForm]
         ]
   }] /; Not[OperatorQ[x]];
Times /: MakeBoxes[Times[x_, y___], StandardForm] :=
   RowBox[{RowBox[{If[Head[x]===Plus || Head[x]===Complex, 
                    Seq["(",MakeBoxes[x,StandardForm],")"], 
                    If[x===-1, "- ",MakeBoxes[x,TraditionalForm]]
                   ]/.Seq->Sequence
         }], 
         Which[Length[{y}] === 0, "",
               Length[{y}] === 1,
                 If[Head[y]===Plus, 
                     Seq["(",MakeBoxes[y, StandardForm],")"],
                     MakeBoxes[y, StandardForm]
                 ]/.Seq->Sequence,
               True, MakeBoxes[Times[y], StandardForm]
         ]
   }] /; Not[OperatorQ[x]];
Protect[Times];
];
(*
PrettyTimes[];
*)
(* Last, to avoid confusions, we prohibit the definition of Times for operators *)

optime = False;

PermitTimes[] := 
 Module[{}, Unprotect[Times];
   If[optime,
     Block[{}, Clear[Times]; PrettyTimes[]],
     Literal[Times[a___]] := 
       Diamond[Times@@DeleteCases[{a},_?OperatorQ], 
                 Diamond@@Select[{a},OperatorQ]] /; 
         Length[{a}] - Length[DeleteCases[{a},_?OperatorQ]] > 1;
   ]  
 ];

PermitTimes /: On[PermitTimes] :=
    Block[{}, 
      optime = True; Message[Times::ison]; PermitTimes[];
    ];

PermitTimes /: Off[PermitTimes] :=
    Block[{},
      optime = False; Message[Times::isoff]; PermitTimes[];
    ];

Times::op =
"Warning -- Attempting to define usual commutative multiplication \
between operators."

Times::isoff =
"Now the usual Times operation between operators has been disabled. \
Any Times expression between operators are automatically turned into Diamond.
Be careful about the results!"

Times::ison =
"Now you can define usual commutative Times operators between operators. \
Be careful about the results!"


Off[Times::isoff];Off[Times::ison];
On[PermitTimes];
(*Off[PermitTimes];*)
On[Times::isoff];On[Times::ison];

(* ------- This line essentially ends the core part of the package ------ *)


(* ---- Shelding the Definition of the new functions in this package ---- *)

Select[{SetOperators, ClearOperators, OperatorQ, SetCommutator, Diamond,
Commutator, CommutatorQ, ClearCommutator, CommutatorToDiamond,
ClearAllOperators,CanonicalOrder,
ListBasicOperators, SetDiamond, ClearDiamond,
OPExpand, SetAllOperators, SetAllOperatorsQ, PermitTimes,
PowerToDiamond,DiamondToPower,CM},SetAttributes[#,ReadProtected]&]; 

(* -----------------------------------------------------------------------*)


(* Contents below this line is in developing *)
(*
expanddiamond = False;
OPExpand /: On[OPExpand] :=
    Module[{},
      expanddiamond = True;
      $Post = ExpandOP;
      ];
OPExpand /: Off[OPExpand] :=
    Module[{},
      expanddiamond = False;
      $Post =.
      ];
ExpandOP[expr_] := If[expanddiamond, OPExpand[expr], expr];
*)
(* End of developing codes *)


End[]

Protect[SetOperators, ClearOperators, OperatorQ, SetCommutator, Diamond,
        Commutator, CommutatorQ, ClearCommutator, CommutatorToDiamond,
        ClearAllOperators,CanonicalOrder,
        ListBasicOperators, SetDiamond, ClearDiamond,
        OPExpand, SetAllOperators, SetAllOperatorsQ, PermitTimes,
        PowerToDiamond,DiamondToPower,CM];
Protect[Minus, Times, Power, Expand];

EndPackage[]

(*SetOperators@@{a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z};*)

(* Functions still need modification:

Off[PermitTimes]

*)


