(*:Mathematica Version: 3.0  *)

(* The DirectProduct Package *)

(* Author:  Liu Zhao 
            Inst. Mod. Phys., Northwest Univ., Xian 710069, China
            lzhao@phy.nwu.edu.cn
            
            Any comments, bug reports, suggestions etc welcomed.
*)

(*:Context: OPLinearAlgebra`OPSolve`  *)

(* Summary:
*)

(* Source: modified LinearEquationsToMatrices function from 
LinearAlgebra`MatrixManipulation` Package into OPLinearEqnToMatrices *)


(*:Keywords: Linear equations, operator-valued matrices, solve
*)

(*:Requirements: OPLinearAlgebra`OPMatrix` *)

(*:Package Version: 1.0 *)

(* History: Created in May, 2000 by Liu Zhao *)

BeginPackage["OPLinearAlgebra`OPSolve`",
             "OPLinearAlgebra`OPScalar`",
             "OPLinearAlgebra`OPMatrix`"
            ];

Unprotect[OPLinearEqnToMatrices, OPSolve]
Clear[OPLinearEqnToMatrices, OPSolve]

OPLinearEqnToMatrices::usage=
"OPLinearEqnToMatrices detects whether an operator-valued linear system of equations \
is of the form A X = B or X A = B. In either cases it would turn the system into \
a coefficient matrix A and a constant vector B (depending whether the system is of \ 
the type A X = B or X A = B, the result will be given in the sequence {A, B} or \
{B, A})"

OPSolve::usage=
"OPSolve[eqns, varlist] solves the system of linear equations eqns with respect \
to varlist. Its functionality is similar to that of the system's function Solve, \
except that it solves operator-valued as well as normal linear equations."

Begin["`Private`"]

oppolydegree[poly_, vars_List] :=
    Module[{j, len=Length[poly], nvars=Length[vars], hed=Head[poly],
      base, expon, deg=0, d, out},
	If [nvars == 0, Return[0]];
	For [j=1, j<=nvars, j++, If [SameQ[poly, vars[[j]]], Return[1]]];
	If [hed===Power || hed===NCPower,
	    If [len=!=2,
	        Message[OPLinearEqnToMatrices::notpoly, poly, vars];
		Return[False]];
	    expon = poly[[2]];
	    j = oppolydegree[expon, vars];
	    If [j=!=0,
	        Message[OPLinearEqnToMatrices::notpoly, poly, vars];
		Return[False]];
	    base = poly[[1]];
	    d = oppolydegree[base, vars];
	    If [d===False,
	        (error = True; Message[ToMatrix::notpoly, poly, vars]; Break)];
	    If [d===0, (deg = 0; Break)];
	    If [!(IntegerQ[expon] && expon>=0),
	        Message[OPLinearEqnToMatrices::notpoly, poly, vars];
		Return[False]];
	    deg = d*expon;
	    ];
	If [hed===Plus,
	    For [j=1, j<=len, j++,
	        d = oppolydegree[poly[[j]], vars];
		If [d===False,
	            Message[OPLinearEqnToMatrices::notpoly, poly, vars];
		    Return[False]];
		deg = Max[deg, d];
		]
	    ];
	If [hed===Times || hed===Diamond,
	    For [j=1, j<=len, j++,
	        d = oppolydegree[poly[[j]], vars];
		If [d===False,
	            Message[OPLinearEqnToMatrices::notpoly, poly, vars];
		    Return[False]];
		deg += d;
		]
	    ];
    deg
    ]

OPLinearEqnToMatrices::noteqn = "`1` is not an equation."
OPLinearEqnToMatrices::notpoly = "`1` is not a polynomial in `2`."
OPLinearEqnToMatrices::notlin = "`1` is not linear in `2`."
OPLinearEqnToMatrices::err = 
"Unknown variables is neither in the left-most nor in the right-most positions, \
cannot treat this type of equations."

Options[OPLinearEqnToMatrices] = {CheckLinearity -> False}

OPLinearEqnToMatrices[oeqns_, vars_List, opts___?OptionQ] :=
	(
	{{},{}}
	) /; Length[oeqns]==0

OPLinearEqnToMatrices[oeqns_, vars_List, opts___?OptionQ] :=
    Module[{eqns=oeqns, len=Length[oeqns], polys, mat, rules, j, poly, eqn,
		chk, giveResult, nvar=Length[Flatten[vars]], i, test, testl, testr, 
		varpos=0, rule, locvars = Flatten[vars]},
		(
	On[CollectPower]; 
		(* this enables the algorithm for determining the coefficient matrix *)
	rules = Thread[locvars -> 0];
	rule[i_]:=Thread[Drop[locvars,{i}] -> 0];
	polys = Map[(#[[1]]-#[[2]])&, eqns];
	vec = polys /. rules;
	polys -= vec;
	testl=OPExpand[Thread[Diamond[locvars[[1]]^(-1), polys]]]/.rule[1];	   
	testr=OPExpand[Thread[Diamond[polys, locvars[[1]]^(-1)]]]/.rule[1];
	If[FreeQ[testl, locvars[[1]]], varpos=-1];
	If[FreeQ[testr, locvars[[1]]], varpos= 1];
  If[varpos==0, Message[OPLinearEqnToMatrices::err]];
  If[Negative[varpos], 
     Block[{},
       For[i=2,i<=nvar,i++,
	       test[i]=OPExpand[Thread[Diamond[locvars[[i]]^(-1), polys]]]/. rule[i];
	       If[Not[FreeQ[test[i],locvars[[i]]]], Message[OPLinearEqnToMatrices::err];
	          varpos=0
	       ]
	     ];
	     test[1]=testl;
	   ],  (* varpos is positive for the next block *)
     Block[{},
       For[i=2,i<=nvar,i++,
	       test[i]=OPExpand[Thread[Diamond[polys, locvars[[i]]^(-1)]]]/. rule[i];
	       If[Not[FreeQ[test[i],locvars[[i]]]], Message[OPLinearEqnToMatrices::err];
	          varpos=0
	       ]
	     ];
	     test[1]=testr;
	   ]
  ];
  mat = Transpose[Table[test[i],{i,1,nvar}]];
  Which[Negative[varpos], Print["Variable Position-> Left"];  {-vec, mat}, 
        Positive[varpos], Print["Variable Position-> Right"]; {mat, -vec},
        varpos==0,        Print["Variable Position-> Mixed"]; {0, 0}
  ]
  ) /; ( chk = TrueQ[CheckLinearity /. {opts} /.
                         Options[OPLinearEqnToMatrices]];
             If [Head[eqns]===And, eqns = Apply[List, eqns]];
             If [Head[eqns]=!=List, eqns = Thread[eqns]];
             If [Head[eqns]=!=List, eqns = {eqns}];
	     giveResult = True;
             If [chk,
                 For[j=1, j<=len, j++,
                     eqn = eqns[[j]];
                     If [Head[eqn]=!=Equal || Length[eqn]=!=2,
                         Message[OPLinearEqnToMatrices::noteqn, eqn];
			 giveResult = False,
                         poly = eqn[[1]] - eqn[[2]];
                         If [giveResult && !PolynomialQ[poly, locvars],
                	     Message[OPLinearEqnToMatrices::notpoly,
				 rhs, locvars];
                             giveResult = False,
                             If [giveResult && oppolydegree[poly, locvars]=!=1,
                                 Message[OPLinearEqnToMatrices::notlin,
					 poly, locvars];
			         giveResult = False
			     ] (* end If *)
			 ] (* end If *)
		     ] (* end If *)	
                 ] (* end For *)
             ]; (* end If *)
	     giveResult 
	   )
	   ]

(* OPSolve routine *)

OPSolve[oeqns_, vars_List] :=
    Module[{tmp,res},
      Print["Calling OPLinearEqnToMatrices ... " ];
      tmp=OPLinearEqnToMatrices[oeqns, Flatten[vars]];
      Print["Calling OPLinearSolve to give the result ... " ];
      res=OPLinearSolve @@ tmp;
      Thread[Flatten[vars] -> res]
    ]


(* ----------------------------------- End ---------------------------------- *)

End[]

Protect[OPLinearEqnToMatrices, OPSolve]

EndPackage[]
