(*: Mathematica Version: 3.0, Not tested for earlier versions. *)

(*: The Poisson Package *)

(* Author:  Liu Zhao 
            Inst. Mod. Phys., Northwest Univ., Xian 710069, China
            lzhao@phy.nwu.edu.cn
            
            Any comments, bug reports, suggestions etc welcomed.
*)

(*: Context: OPLinearAlgebra`Poisson`  *)

(*: Summary: A Poisson bracket evaluation package.
*)

(*: Keywords: Poisson bracket, Canonical structure
*)

(*: Requirements: OPLinearAlgebra`OPScalar`  *)

(*: Package Version: 1.0 *)

(*: History: Created in May, 2000 by Liu Zhao *)

BeginPackage["OPLinearAlgebra`Poisson`",
             "OPLinearAlgebra`OPScalar`",
             "LinearAlgebra`MatrixManipulation`"];

Unprotect[Poisson, SetPoisson, PB, ShowCoordinates, ShowMomenta, ShowCanonical,
          DefineTheory, SetCanonical, ShowTheory, CirclePoisson, ClearPoisson];

Clear[Poisson, SetPoisson, PB, ShowCoordinates, ShowMomenta, ShowCanonical,
      DefineTheory, SetCanonical, ShowTheory, CirclePoisson, ClearPoisson];

(* Brief help for new functions *)

Poisson::usage=
"Poisson[x, y] gives the Poisson bracket of x, y. Poisson[x,y] cannot be assigned
directly, any assignment must be done through SetPoisson."

SetPoisson::usage=
"SetPoisson[x, y, z] sets the Poisson bracket of x and y to z."

ClearPoisson::usage=
"ClearPoisson[x,y] clears the assigned value for PB[x,y]."

PB::usage=
"PB is the output form of Poisson. Do not use this for any purpose."

CirclePoisson::usage=
"CirclePoisson[x, y] is an analogue of CircleCommutator for Poisson brackets. See \
CircleCommutator for detailes."


ShowCoordinates::usage=
"ShowCoordinates[] gives the list of all canonical coordinates variables."

ShowMomenta::usage=
"ShowMomenta[] gives the list of all canonical momenta."

ShowCanonical::usage=
"ShowCanonical[] returns a matrix, of which the first row is the list of canonical \
coordinates, and the second row is the list of canonical momenta."

DefineTheory::usage=
"DefineTheory[x, y] declears the system being studied is whether Classical Mechanics, \
Quantum Mechanic, Classical Field, or Quantum Field."

SetCanonical::usage=
"SetCanonical[{x,y}..] defines all the canonical variables as well as their \
canonical stucture, depending which kind the theory is."

ShowTheory::usage=
"ShowTheory[] indicates the currently defined type of theory."


Begin["`Private`"]

Clear[pb];
Off[Times::ison];Off[Times::op];
On[PermitTimes];

(* ----- Contents below is a developing part of the package on Poisson Brackets ---- *)

(* Version of Poisson bracket *)

PB[y_, fn_[z___]] := 
    Module[{var, l = Length[{z}],ret}, 
      SetOperators[var];
      ret=Sum[(D[fn[z] /. {{z}[[i]] -> var}, var] PB[y, var] /. 
         {var -> {z}[[i]]}), {i, 1, l}];
      ClearOperators[var];
      Return[ret]
    ]/; Not[OperatorQ[fn]] && fn=!=PB;
    
PB[fn_[z___], y_] := 
    Module[{var, l = Length[{z}],ret}, 
      SetOperators[var];
      ret=Sum[(D[fn[z] /. {{z}[[i]] -> var}, var] PB[var, y] /. 
         {var -> {z}[[i]]}), {i, 1, l}];
      ClearOperators[var];
      Return[ret]
    ]/; Not[OperatorQ[fn]] && fn=!=PB;
    
(*
PB[x_ y_, z_]:=PB[x, z] y + x PB[y, z];
PB[z_, x_ y_]:=PB[z, x] y + x PB[z, y];

PB[x_+y_, z_]:=PB[x, z] + PB[y, z];
PB[z_, x_+y_]:=PB[z, x] + PB[z, y];

PB[PB[a_, b_], c_]:= -PB[c, PB[a, b]] /; Head[c]=!=PB;

PB[a_, PB[b_, c_]] :=
  - PB[c, PB[a, b]] + PB[b, PB[a, c]] /;
      And[OrderedQ[{b,a}], OrderedQ[{c,a}], Head[PB[b, a]]===PB,
             Head[PB[c, a]]===PB
      ];

PB[PB[a_, b_], PB[c_, d_]]:= - PB[b, PB[a, PB[c, d]]] + PB[a, PB[b, PB[c, d]]];
*)

PB[x_, y_] := 0 /; Not[OperatorQ[x]] || Not[OperatorQ[y]] || x===y ;

PB[y_,x_]:=-PB[x,y]/; OrderedQ[{x,y}];

Poisson[x_, y_] := 
   Module[{tmp1,tmp2},
     tmp1=Expand[CM[x, y] //. {Diamond->Times, CM->PB}];
     tmp2=Expand[tmp1 /.PB->CM];
     Return[Expand[tmp2/.CM->PB]]
   ];


SetPoisson[x_,y_,z_]:= 
    Module[{},Unprotect[PB];
      If[OrderedQ[{x,y}], PB[x,y]:=z, PB[y,x]:=-z];
      Protect[PB];
    ]/; OperatorQ[x] && OperatorQ[y]

ClearPoisson[x_,y_]:= 
    Module[{},Unprotect[PB];
      If[OrderedQ[{x,y}], PB[x,y]=.;, PB[y,x]=.;];
      Protect[PB];
    ];
    


CirclePoisson[x_?MatrixQ,y_?MatrixQ] := BlockMatrix[Outer[Poisson,x,y]];

(* Formatting output for Poisson *)

PB/: MakeBoxes[PB[x_,y_],TraditionalForm]:=
  RowBox[{"{",
          MakeBoxes[x,TraditionalForm],",",MakeBoxes[y,TraditionalForm],
          SubscriptBox["}","PB"]
  }];
PB/: MakeBoxes[PB[x_,y_],StandardForm]:=
  RowBox[{"{",
          MakeBoxes[x,StandardForm],",",MakeBoxes[y,StandardForm],
          SubscriptBox["}","PB"]
  }];


(* Codes for using Commutator and Poisson in physical theories *)

theory=.;
Coordinates=.;
Momenta=.;

DefineTheory[x_,y_] :=
    If[(ToString[x] === "Classical" ) || ( ToString[x] === "Quantum"),
        If[(ToString[y] === "Mechanics") || (ToString[y] === "Field"),
           theory = {x, y}, Message[DefineTheory::theory]
        ]
    ];

DefineTheory::theory = 
"Theory has to be one of {Classical, Machanics}, {Quantum, Mechanics},
{Classical, Field} or {Quamtum, Field}. Failed."

SetCanonical[x : {_, _} ..] :=
   If[Not[MemberQ[{"Classical", "Quantum"},ToString[theory[[1]]]]] ||
             Not[MemberQ[{"Mechanics", "Field"}, ToString[theory[[2]]]]],
      Message[SetCanonical::theory],
      Module[{v1, y, tmp, dim},
             v1[y : {_, _} ..] := Transpose[{y}];
             tmp = v1[x]; Coordinates = tmp[[1]]; Momenta = tmp[[2]];
             dim=Length[Coordinates];
             SetOperators[Coordinates]; SetOperators[Momenta];
             Which[ToString[theory[[1]]]==="Classical" &&
                   ToString[theory[[2]]]==="Mechanics",
                       Do[SetPoisson[Coordinates[[i]],Momenta[[j]], If[i==j,1,0]],
                                 {i,1,dim},{j,1,dim}],
                   ToString[theory[[1]]]==="Quantum" &&
                   ToString[theory[[2]]]==="Mechanics",
                       Do[SetCommutator[Coordinates[[i]],Momenta[[j]],If[i==j,1,0]],
                                 {i,1,dim},{j,1,dim}],
                   ToString[theory[[1]]]==="Classical" &&
                   ToString[theory[[2]]]==="Field",
                       Do[SetPoisson[Coordinates[[i]],Momenta[[j]],DiracDelta[i,j]],
                                 {i,1,dim},{j,1,dim}],
                   ToString[theory[[1]]]==="Quantum" &&
                   ToString[theory[[2]]]==="Field",
                       Do[SetCommutator[Coordinates[[i]],Momenta[[j]],DiracDelta[i,j]],
                                 {i,1,dim},{j,1,dim}]
             ]
      ]
   ];

SetCanonical::theory=
"Theory not found! Has to define theory first. (Use DefineTheory please)"

ShowCoordinates[]:=SequenceForm["Coordinates = ", Coordinates];
ShowMomenta[]:=SequenceForm["Momenta = ", Momenta];
ShowCanonical[]:={Coordinates,Momenta};
ShowTheory[]:=
   If[ToString[Head[theory]]!="List",
      Message[SetCanonical::theory],
      StringJoin[ToString[theory[[1]]]," ",ToString[theory[[2]]]]
   ];


(* ----------------------------------------------------------------- *)
Select[{Poisson,SetPoisson,PB,ShowCoordinates,ShowMomenta,ShowCanonical,
  DefineTheory,SetCanonical,ShowTheory,CirclePoisson, ClearPoisson},
  SetAttributes[#,ReadProtected]&];
(* ----------------------------------------------------------------- *)

End[]

Protect[Poisson,SetPoisson,PB,ShowCoordinates,ShowMomenta,ShowCanonical,
        DefineTheory,SetCanonical,ShowTheory,CirclePoisson, ClearPoisson];
Protect[Minus, Times, Power, Expand];

EndPackage[]
