(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      8830,        284]*)
(*NotebookOutlinePosition[      9482,        307]*)
(*  CellTagsIndexPosition[      9438,        303]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Chevalley basis of simple Lie Algebras", "Title"],

Cell[CellGroupData[{

Cell[TextData[StyleBox["First we load the package OPScalar.m and define the \
Chevalley generating relations\nfor simple Lie algebras", "Text"]], "Section"],

Cell["This loads the package:", "Text"],

Cell[BoxData[
    \(<< OPLinearAlgebra`OPScalar`\)], "Input",
  GeneratedCell->True],

Cell["\<\
Now define the Chevalley generators with Head {e, f, h} to be Operators:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(SetOperators[e, f, h]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`{e, f, h}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Since K[i,j] is undefined, it is treated as ordinary variables in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", rather than Operators. These play the role of Cartan matrix elements of \
the Lie algebra being considered. The next 4 lines define the Chevalley \
generating relations:"
}], "Text"],

Cell[BoxData[
    \(SetCommutator[h[i_], e[j_], K[i, j] e[j]]\)], "Input"],

Cell[BoxData[
    \(SetCommutator[h[i_], f[j_], \(-K[i, j]\) f[j]]\)], "Input"],

Cell[BoxData[
    \(SetCommutator[e[i_], f[j_], If[i - j \[Equal] 0, h[i], 0]]\)], "Input"],

Cell[BoxData[
    \(SetCommutator[h[i_], h[j_], 0]\)], "Input"],

Cell["Now we can check whether the definition works as expected:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[e[1], f[1]]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`h(1)\)], "Output"]
}, Open  ]],

Cell["OK, that finishes the definition.", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Lie multiplications for root vectors corresponding to roots of heights from 2 \
to 4:\
\>", "Section"],

Cell["\<\
First we define the root vectors for roots of heights ranging from 2 to 4,\
\>", "Text"],

Cell[BoxData[
    \(E2[i_] := Commutator[e[i], e[i + 1]]\)], "Input"],

Cell[BoxData[
    \(F2[i_] := Commutator[f[i], f[i + 1]]\)], "Input"],

Cell[BoxData[
    \(E3[i_] := Commutator[e[i], Commutator[e[i + 1], e[i + 2]]]\)], "Input"],

Cell[BoxData[
    \(F3[i_] := Commutator[f[i], Commutator[f[i + 1], f[i + 2]]]\)], "Input"],

Cell[BoxData[
    \(E4[i_] := 
      Commutator[e[i], 
        Commutator[e[i + 1], Commutator[e[i + 2], e[i + 3]]]]\)], "Input"],

Cell[BoxData[
    \(F4[i_] := 
      Commutator[f[i], 
        Commutator[f[i + 1], Commutator[f[i + 2], f[i + 3]]]]\)], "Input"],

Cell["\<\
Now the Lie product between E2[i] and F2[j] can be easily evaluated,\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(H2[i] = Commutator[E2[i], F2[i]]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`K[i, i + 1] \(h(i + 1)\) + 
      K[i + 1, i] \(h(i)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[E2[i], F2[i + 1]]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`0\)], "Output"]
}, Open  ]],

Cell["Similar calculations for E3[i] and F3[j] can be persued,", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(H3[i] = Commutator[E3[i], F3[i]] // Simplify\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`\((K[i, i + 1] + K[i, i + 2])\) \(K[i + 1, i + 2] \(h(
            i + 2)\)\) + \((K[i, i + 1] + K[i, i + 2])\) \(K[i + 2, 
            i + 1] \(h(i + 1)\)\) + \((K[i + 1, i + 2] K[i + 2, i] + 
            K[i + 1, i] K[i + 2, i + 1])\) \(h(i)\)\)], "Output"]
}, Open  ]],

Cell["And this is a big one for E4[i] and F4[i], ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(H4[i] = Commutator[E4[i], F4[i]] // OPExpand\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`K[i, 
          i + 1] \(K[i + 1, i + 2] \(K[i + 2, i + 3] \(h(i + 3)\)\)\) + 
      K[i, i + 2] \(K[i + 1, i + 2] \(K[i + 2, i + 3] \(h(i + 3)\)\)\) + 
      K[i, i + 3] \(K[i + 1, i + 2] \(K[i + 2, i + 3] \(h(i + 3)\)\)\) + 
      K[i, i + 1] \(K[i + 1, i + 3] \(K[i + 2, i + 3] \(h(i + 3)\)\)\) + 
      K[i, i + 2] \(K[i + 1, i + 3] \(K[i + 2, i + 3] \(h(i + 3)\)\)\) + 
      K[i, i + 3] \(K[i + 1, i + 3] \(K[i + 2, i + 3] \(h(i + 3)\)\)\) + 
      K[i + 1, i + 2] \(K[i + 3, i] \(K[i + 2, i + 3] \(h(i)\)\)\) + 
      K[i + 1, i + 3] \(K[i + 3, i] \(K[i + 2, i + 3] \(h(i)\)\)\) + 
      K[i, i + 1] \(K[i + 3, i + 1] \(K[i + 2, i + 3] \(h(i + 1)\)\)\) + 
      K[i, i + 2] \(K[i + 3, i + 1] \(K[i + 2, i + 3] \(h(i + 1)\)\)\) + 
      K[i, i + 3] \(K[i + 3, i + 1] \(K[i + 2, i + 3] \(h(i + 1)\)\)\) + 
      K[i + 1, i] \(K[i + 3, i + 1] \(K[i + 2, i + 3] \(h(i)\)\)\) + 
      K[i, i + 1] \(K[i + 1, i + 2] \(K[i + 3, i + 2] \(h(i + 2)\)\)\) + 
      K[i, i + 2] \(K[i + 1, i + 2] \(K[i + 3, i + 2] \(h(i + 2)\)\)\) + 
      K[i, i + 3] \(K[i + 1, i + 2] \(K[i + 3, i + 2] \(h(i + 2)\)\)\) + 
      K[i, i + 1] \(K[i + 1, i + 3] \(K[i + 3, i + 2] \(h(i + 2)\)\)\) + 
      K[i, i + 2] \(K[i + 1, i + 3] \(K[i + 3, i + 2] \(h(i + 2)\)\)\) + 
      K[i, i + 3] \(K[i + 1, i + 3] \(K[i + 3, i + 2] \(h(i + 2)\)\)\) + 
      K[i + 1, i + 2] \(K[i + 2, i] \(K[i + 3, i + 2] \(h(i)\)\)\) + 
      K[i + 1, i + 3] \(K[i + 2, i] \(K[i + 3, i + 2] \(h(i)\)\)\) + 
      K[i, i + 1] \(K[i + 2, i + 1] \(K[i + 3, i + 2] \(h(i + 1)\)\)\) + 
      K[i, i + 2] \(K[i + 2, i + 1] \(K[i + 3, i + 2] \(h(i + 1)\)\)\) + 
      K[i, i + 3] \(K[i + 2, i + 1] \(K[i + 3, i + 2] \(h(i + 1)\)\)\) + 
      K[i + 1, i] \(K[i + 2, 
            i + 1] \(K[i + 3, i + 2] \(h(i)\)\)\)\)], "Output"]
}, Open  ]],

Cell["\<\
This is another example, the Lie product between h[i] and F4[j],\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[h[i], F4[j]] // OPExpand\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`\(\(-\)\(\ \)\) \(K[i, j] \(\([\)\(f(
            j)\)\(,\)\(\([\)\(f(j + 1)\)\(,\)\(\([\)\(f(j + 2)\)\(,\)\(f(
            j + 3)\)\(]\)\)\(]\)\)\(]\)\)\) - 
      K[i, j + 1] \(\([\)\(f(j)\)\(,\)\(\([\)\(f(j + 1)\)\(,\)\(\([\)\(f(
          j + 2)\)\(,\)\(f(j + 3)\)\(]\)\)\(]\)\)\(]\)\) - 
      K[i, j + 2] \(\([\)\(f(j)\)\(,\)\(\([\)\(f(j + 1)\)\(,\)\(\([\)\(f(
          j + 2)\)\(,\)\(f(j + 3)\)\(]\)\)\(]\)\)\(]\)\) - 
      K[i, j + 3] \(\([\)\(f(j)\)\(,\)\(\([\)\(f(j + 1)\)\(,\)\(\([\)\(f(
          j + 2)\)\(,\)\(f(j + 3)\)\(]\)\)\(]\)\)\(]\)\)\)], "Output"]
}, Open  ]],

Cell["\<\
So far, the Cartan matrix elements K[i,j] are not specified. Now we provide \
the Cartan matrix for A series Lie algebras,\
\>", "Text"],

Cell[BoxData[
    \(K[i_, j_] := 
      Which[i - j \[Equal] 0, 2, i - j == 1, \(-1\), 
        i - j \[Equal] \(-1\), \(-1\), True, 0]\)], "Input"],

Cell["\<\
Now let us check the earlier result for the specific case of A type Lie \
algebras:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(H2[i]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`\(\(-\)\(\ \)\) \(h(i)\) - h(i + 1)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(H3[i]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`h(i) + h(i + 1) + h(i + 2)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(H4[i]\)], "Input"],

Cell[BoxData[
    \(TraditionalForm\`\(\(-\)\(\ \)\) \(h(i)\) - h(i + 1) - h(i + 2) - 
      h(i + 3)\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 991}, {0, 695}},
WindowSize->{710, 575},
WindowMargins->{{0, Automatic}, {Automatic, 0}}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 55, 0, 105, "Title"],

Cell[CellGroupData[{
Cell[1819, 55, 156, 1, 78, "Section"],
Cell[1978, 58, 39, 0, 33, "Text"],
Cell[2020, 60, 84, 2, 30, "Input"],
Cell[2107, 64, 96, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[2228, 70, 54, 1, 30, "Input"],
Cell[2285, 73, 60, 1, 29, "Output"]
}, Open  ]],
Cell[2360, 77, 327, 7, 52, "Text"],
Cell[2690, 86, 74, 1, 30, "Input"],
Cell[2767, 89, 79, 1, 30, "Input"],
Cell[2849, 92, 91, 1, 30, "Input"],
Cell[2943, 95, 63, 1, 30, "Input"],
Cell[3009, 98, 74, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[3108, 102, 55, 1, 30, "Input"],
Cell[3166, 105, 55, 1, 29, "Output"]
}, Open  ]],
Cell[3236, 109, 49, 0, 33, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3322, 114, 112, 3, 53, "Section"],
Cell[3437, 119, 98, 2, 33, "Text"],
Cell[3538, 123, 69, 1, 30, "Input"],
Cell[3610, 126, 69, 1, 30, "Input"],
Cell[3682, 129, 91, 1, 30, "Input"],
Cell[3776, 132, 91, 1, 30, "Input"],
Cell[3870, 135, 129, 3, 30, "Input"],
Cell[4002, 140, 129, 3, 30, "Input"],
Cell[4134, 145, 92, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[4251, 151, 65, 1, 30, "Input"],
Cell[4319, 154, 105, 2, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4461, 161, 61, 1, 30, "Input"],
Cell[4525, 164, 52, 1, 29, "Output"]
}, Open  ]],
Cell[4592, 168, 72, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[4689, 172, 77, 1, 30, "Input"],
Cell[4769, 175, 296, 4, 48, "Output"]
}, Open  ]],
Cell[5080, 182, 59, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[5164, 186, 77, 1, 30, "Input"],
Cell[5244, 189, 1817, 26, 238, "Output"]
}, Open  ]],
Cell[7076, 218, 88, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[7189, 224, 68, 1, 30, "Input"],
Cell[7260, 227, 608, 9, 48, "Output"]
}, Open  ]],
Cell[7883, 239, 146, 3, 33, "Text"],
Cell[8032, 244, 148, 3, 30, "Input"],
Cell[8183, 249, 107, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[8315, 256, 38, 1, 30, "Input"],
Cell[8356, 259, 86, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8479, 265, 38, 1, 30, "Input"],
Cell[8520, 268, 77, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8634, 274, 38, 1, 30, "Input"],
Cell[8675, 277, 115, 2, 29, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

